"""Network Interface Management

.. warning:: This module is disabled in 6.x and will removed in 7.x. Please use networking
             libraries instead.

This module provides a registry of configured NICs.
It is used by the 'socket' module to look up a suitable
NIC when a socket is created."""

from __future__ import annotations

from typing import List

def route() -> List[object]:
    """Returns a list of all configured NICs."""
    ...



"""Support for WizNet hardware, including the WizNet 5500 Ethernet adaptor.


.. warning:: This module is disabled in 6.x and will removed in 7.x. Please use networking
             libraries instead.
"""

from __future__ import annotations

from typing import Optional, Tuple

import busio
import microcontroller

class WIZNET5K:
    """Wrapper for Wiznet 5500 Ethernet interface"""

    def __init__(
        self,
        spi: busio.SPI,
        cs: microcontroller.Pin,
        rst: microcontroller.Pin,
        dhcp: bool = True,
    ) -> None:
        """Create a new WIZNET5500 interface using the specified pins

        :param ~busio.SPI spi: spi bus to use
        :param ~microcontroller.Pin cs: pin to use for Chip Select
        :param ~microcontroller.Pin rst: pin to use for Reset (optional)
        :param bool dhcp: boolean flag, whether to start DHCP automatically (optional, keyword only, default True)

        * The reset pin is optional: if supplied it is used to reset the
          wiznet board before initialization.
        * The SPI bus will be initialized appropriately by this library.
        * At present, the WIZNET5K object is a singleton, so only one WizNet
          interface is supported at a time."""
        ...
    connected: bool
    """(boolean, readonly) is this device physically connected?"""

    dhcp: bool
    """(boolean, readwrite) is DHCP active on this device?

    * set to True to activate DHCP, False to turn it off"""
    def ifconfig(
        self, params: Optional[Tuple[str, str, str, str]] = None
    ) -> Optional[Tuple[str, str, str, str]]:
        """Called without parameters, returns a tuple of
        (ip_address, subnet_mask, gateway_address, dns_server)

        Or can be called with the same tuple to set those parameters.
        Setting ifconfig parameters turns DHCP off, if it was on."""
        ...



"""Frequency-domain functions"""

from __future__ import annotations

from typing import Optional, Tuple

import ulab

def fft(
    r: ulab.ndarray, c: Optional[ulab.ndarray] = None
) -> Tuple[ulab.ndarray, ulab.ndarray]:
    """
    :param ulab.ndarray r: A 1-dimension array of values whose size is a power of 2
    :param ulab.ndarray c: An optional 1-dimension array of values whose size is a power of 2, giving the complex part of the value
    :return tuple (r, c): The real and complex parts of the FFT

    Perform a Fast Fourier Transform from the time domain into the frequency domain

    See also ~ulab.extras.spectrum, which computes the magnitude of the fft,
    rather than separately returning its real and imaginary parts."""
    ...

def ifft(
    r: ulab.ndarray, c: Optional[ulab.ndarray] = None
) -> Tuple[ulab.ndarray, ulab.ndarray]:
    """
    :param ulab.ndarray r: A 1-dimension array of values whose size is a power of 2
    :param ulab.ndarray c: An optional 1-dimension array of values whose size is a power of 2, giving the complex part of the value
    :return tuple (r, c): The real and complex parts of the inverse FFT

    Perform an Inverse Fast Fourier Transform from the frequeny domain into the time domain"""
    ...



"""Tracks button presses read through a shift register"""

from __future__ import annotations

import digitalio

class GamePadShift:
    """Scan buttons for presses through a shift register"""

    def __init__(
        self,
        clock: digitalio.DigitalInOut,
        data: digitalio.DigitalInOut,
        latch: digitalio.DigitalInOut,
    ) -> None:
        """Initializes button scanning routines.

        The ``clock``, ``data`` and ``latch`` parameters are ``DigitalInOut``
        objects connected to the shift register controlling the buttons.

        The button presses are accumulated, until the ``get_pressed`` method
        is called, at which point the button state is cleared, and the new
        button presses start to be recorded.

        Only one gamepad (`gamepad.GamePad` or `gamepadshift.GamePadShift`)
        may be used at a time."""
        ...
    def get_pressed(self) -> int:
        """Get the status of buttons pressed since the last call and clear it.

        Returns an 8-bit number, with bits that correspond to buttons,
        which have been pressed (or held down) since the last call to this
        function set to 1, and the remaining bits set to 0. Then it clears
        the button state, so that new button presses (or buttons that are
        held down) can be recorded for the next call."""
        ...
    def deinit(self) -> None:
        """Disable button scanning."""
        ...



"""Lightweight 2d shapes for displays"""

from __future__ import annotations

from typing import List, Tuple, Union

import displayio

class Circle:
    def __init__(self, radius: int) -> None:
        """Circle is positioned on screen by its center point.

        :param radius: The radius of the circle in pixels"""
    radius: int
    """The radius of the circle in pixels."""

class Polygon:
    def __init__(self, points: List[Tuple[int, int]]) -> None:
        """Represents a closed shape by ordered vertices

        :param points: Vertices for the polygon"""
    points: List[Tuple[int, int]]
    """Set a new look and shape for this polygon"""

class Rectangle:
    def __init__(self, width: int, height: int) -> None:
        """Represents a rectangle by defining its bounds

        :param width: The number of pixels wide
        :param height: The number of pixels high"""

class VectorShape:
    def __init__(
        self,
        shape: Union[Polygon, Rectangle, Circle],
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        x: int = 0,
        y: int = 0,
    ) -> None:
        """Binds a vector shape to a location and pixel color

        :param shape: The shape to draw.
        :param pixel_shader: The pixel shader that produces colors from values
        :param x: Initial x position of the center axis of the shape within the parent.
        :param y: Initial y position of the center axis of the shape within the parent."""
        ...
    x: int
    """X position of the center point of the shape in the parent."""

    y: int
    """Y position of the center point of the shape in the parent."""

    pixel_shader: Union[displayio.ColorConverter, displayio.Palette]
    """The pixel shader of the shape."""



"""Numerical approximation methods"""

from __future__ import annotations

from typing import Optional

import ulab

def interp(
    x: ulab.ndarray,
    xp: ulab.ndarray,
    fp: ulab.ndarray,
    *,
    left: Optional[float] = None,
    right: Optional[float] = None
) -> ulab.ndarray:
    """
    :param ulab.ndarray x: The x-coordinates at which to evaluate the interpolated values.
    :param ulab.ndarray xp: The x-coordinates of the data points, must be increasing
    :param ulab.ndarray fp: The y-coordinates of the data points, same length as xp
    :param left: Value to return for ``x < xp[0]``, default is ``fp[0]``.
    :param right: Value to return for ``x > xp[-1]``, default is ``fp[-1]``.

    Returns the one-dimensional piecewise linear interpolant to a function with given discrete data points (xp, fp), evaluated at x."""
    ...

def trapz(y: ulab.ndarray, x: Optional[ulab.ndarray] = None, dx: float = 1.0) -> float:
    """
    :param 1D ulab.ndarray y: the values of the dependent variable
    :param 1D ulab.ndarray x: optional, the coordinates of the independent variable. Defaults to uniformly spaced values.
    :param float dx: the spacing between sample points, if x=None

    Returns the integral of y(x) using the trapezoidal rule.
    """
    ...



"""Element-by-element functions

These functions can operate on numbers, 1-D iterables, and arrays of 1 to 4 dimensions by
applying the function to every element in the array.  This is typically
much more efficient than expressing the same operation as a Python loop."""

from __future__ import annotations

from typing import Callable, Optional, Union

import ulab
from ulab import _ArrayLike, _DType

def acos(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse cosine function"""
    ...

def acosh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse hyperbolic cosine function"""
    ...

def asin(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse sine function"""
    ...

def asinh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse hyperbolic sine function"""
    ...

def around(a: _ArrayLike, *, decimals: int = 0) -> ulab.ndarray:
    """Returns a new float array in which each element is rounded to
    ``decimals`` places."""
    ...

def atan(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse tangent function; the return values are in the
    range [-pi/2,pi/2]."""
    ...

def arctan2(ya: _ArrayLike, xa: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse tangent function of y/x; the return values are in
    the range [-pi, pi]."""
    ...

def atanh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the inverse hyperbolic tangent function"""
    ...

def ceil(a: _ArrayLike) -> ulab.ndarray:
    """Rounds numbers up to the next whole number"""
    ...

def cos(a: _ArrayLike) -> ulab.ndarray:
    """Computes the cosine function"""
    ...

def cosh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the hyperbolic cosine function"""
    ...

def degrees(a: _ArrayLike) -> ulab.ndarray:
    """Converts angles from radians to degrees"""
    ...

def erf(a: _ArrayLike) -> ulab.ndarray:
    """Computes the error function, which has applications in statistics"""
    ...

def erfc(a: _ArrayLike) -> ulab.ndarray:
    """Computes the complementary error function, which has applications in statistics"""
    ...

def exp(a: _ArrayLike) -> ulab.ndarray:
    """Computes the exponent function."""
    ...

def expm1(a: _ArrayLike) -> ulab.ndarray:
    """Computes $e^x-1$.  In certain applications, using this function preserves numeric accuracy better than the `exp` function."""
    ...

def floor(a: _ArrayLike) -> ulab.ndarray:
    """Rounds numbers up to the next whole number"""
    ...

def gamma(a: _ArrayLike) -> ulab.ndarray:
    """Computes the gamma function"""
    ...

def lgamma(a: _ArrayLike) -> ulab.ndarray:
    """Computes the natural log of the gamma function"""
    ...

def log(a: _ArrayLike) -> ulab.ndarray:
    """Computes the natural log"""
    ...

def log10(a: _ArrayLike) -> ulab.ndarray:
    """Computes the log base 10"""
    ...

def log2(a: _ArrayLike) -> ulab.ndarray:
    """Computes the log base 2"""
    ...

def radians(a: _ArrayLike) -> ulab.ndarray:
    """Converts angles from degrees to radians"""
    ...

def sin(a: _ArrayLike) -> ulab.ndarray:
    """Computes the sine function"""
    ...

def sinh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the hyperbolic sine"""
    ...

def sqrt(a: _ArrayLike) -> ulab.ndarray:
    """Computes the square root"""
    ...

def tan(a: _ArrayLike) -> ulab.ndarray:
    """Computes the tangent"""
    ...

def tanh(a: _ArrayLike) -> ulab.ndarray:
    """Computes the hyperbolic tangent"""
    ...

def vectorize(
    f: Union[Callable[[int], float], Callable[[float], float]],
    *,
    otypes: Optional[_DType] = None
) -> Callable[[_ArrayLike], ulab.ndarray]:
    """
    :param callable f: The function to wrap
    :param otypes: List of array types that may be returned by the function.  None is interpreted to mean the return value is float.

    Wrap a Python function ``f`` so that it can be applied to arrays.
    The callable must return only values of the types specified by ``otypes``, or the result is undefined."""
    ...



from __future__ import annotations

import ulab

def spectrogram(r: ulab.ndarray) -> ulab.ndarray:
    """
    :param ulab.ndarray r: A 1-dimension array of values whose size is a power of 2

    Computes the spectrum of the input signal.  This is the absolute value of the (complex-valued) fft of the signal.
    This function is similar to scipy's ``scipy.signal.spectrogram``."""
    ...



from __future__ import annotations

import ulab

def trace(m: ulab.numpy.ndarray) -> float:
    """
    :param m: a square matrix

    Compute the trace of the matrix, the sum of its diagonal elements."""
    ...



"""pseudo-random numbers and choices

The `random` module is a strict subset of the CPython `cpython:random`
module. So, code written in CircuitPython will work in CPython but not
necessarily the other way around.

Like its CPython cousin, CircuitPython's random seeds itself on first use
with a true random from os.urandom() when available or the uptime otherwise.
Once seeded, it will be deterministic, which is why its bad for cryptography.

.. warning:: Numbers from this module are not cryptographically strong! Use
  bytes from `os.urandom` directly for true randomness."""

from __future__ import annotations

from typing import Sequence, Tuple, TypeVar

_T = TypeVar("_T")

def seed(seed: int) -> None:
    """Sets the starting seed of the random  number generation. Further calls to
    `random` will return deterministic results afterwards."""
    ...

def getrandbits(k: int) -> int:
    """Returns an integer with *k* random bits."""
    ...

def randrange(stop: Tuple[int, int, int]) -> int:
    """Returns a randomly selected integer from ``range(start, stop, step)``."""
    ...

def randint(a: int, b: int) -> int:
    """Returns a randomly selected integer between a and b inclusive. Equivalent
    to ``randrange(a, b + 1, 1)``"""
    ...

def choice(seq: Sequence[_T]) -> _T:
    """Returns a randomly selected element from the given sequence. Raises
    IndexError when the sequence is empty."""
    ...

def random() -> float:
    """Returns a random float between 0 and 1.0."""
    ...

def uniform(a: float, b: float) -> float:
    """Returns a random float between a and b. It may or may not be inclusive
    depending on float rounding."""
    ...



from __future__ import annotations

from typing import Tuple

import ulab
import ulab.numpy

"""Linear algebra functions"""

def cholesky(A: ulab.numpy.ndarray) -> ulab.numpy.ndarray:
    """
    :param ~ulab.numpy.ndarray A: a positive definite, symmetric square matrix
    :return ~ulab.numpy.ndarray L: a square root matrix in the lower triangular form
    :raises ValueError: If the input does not fulfill the necessary conditions

    The returned matrix satisfies the equation m=LL*"""
    ...

def det(m: ulab.numpy.ndarray) -> float:
    """
    :param: m, a square matrix
    :return float: The determinant of the matrix

    Computes the eigenvalues and eigenvectors of a square matrix"""
    ...

def eig(m: ulab.numpy.ndarray) -> Tuple[ulab.numpy.ndarray, ulab.numpy.ndarray]:
    """
    :param m: a square matrix
    :return tuple (eigenvectors, eigenvalues):

    Computes the eigenvalues and eigenvectors of a square matrix"""
    ...

def inv(m: ulab.numpy.ndarray) -> ulab.numpy.ndarray:
    """
    :param ~ulab.numpy.ndarray m: a square matrix
    :return: The inverse of the matrix, if it exists
    :raises ValueError: if the matrix is not invertible

    Computes the inverse of a square matrix"""
    ...

def norm(x: ulab.numpy.ndarray) -> float:
    """
    :param ~ulab.numpy.ndarray x: a vector or a matrix

    Computes the 2-norm of a vector or a matrix, i.e., ``sqrt(sum(x*x))``, however, without the RAM overhead."""
    ...



"""Digital protocols implemented by the CPU

The `bitbangio` module contains classes to provide digital bus protocol
support regardless of whether the underlying hardware exists to use the
protocol.

First try to use `busio` module instead which may utilize peripheral
hardware to implement the protocols. Native implementations will be faster
than bitbanged versions and have more capabilities.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info.

For example::

  import bitbangio
  from board import *

  i2c = bitbangio.I2C(SCL, SDA)
  print(i2c.scan())
  i2c.deinit()

This example will initialize the the device, run
:py:meth:`~bitbangio.I2C.scan` and then :py:meth:`~bitbangio.I2C.deinit` the
hardware. The last step is optional because CircuitPython automatically
resets hardware after a program finishes."""

from __future__ import annotations

from typing import List, Optional

import microcontroller
from _typing import ReadableBuffer, WriteableBuffer

class I2C:
    """Two wire serial protocol"""

    def __init__(
        self,
        scl: microcontroller.Pin,
        sda: microcontroller.Pin,
        *,
        frequency: int = 400000,
        timeout: int = 255
    ) -> None:
        """I2C is a two-wire protocol for communicating between devices.  At the
        physical level it consists of 2 wires: SCL and SDA, the clock and data
        lines respectively.

        .. seealso:: Using this class directly requires careful lock management.
            Instead, use :class:`~adafruit_bus_device.i2c_device.I2CDevice` to
            manage locks.

        .. seealso:: Using this class to directly read registers requires manual
            bit unpacking. Instead, use an existing driver or make one with
            :ref:`Register <register-module-reference>` data descriptors.

        :param ~microcontroller.Pin scl: The clock pin
        :param ~microcontroller.Pin sda: The data pin
        :param int frequency: The clock frequency of the bus
        :param int timeout: The maximum clock stretching timeout in microseconds"""
        ...
    def deinit(self) -> None:
        """Releases control of the underlying hardware so other classes can use it."""
        ...
    def __enter__(self) -> I2C:
        """No-op used in Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware on context exit. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def scan(self) -> List[int]:
        """Scan all I2C addresses between 0x08 and 0x77 inclusive and return a list of
        those that respond.  A device responds if it pulls the SDA line low after
        its address (including a read bit) is sent on the bus."""
        ...
    def try_lock(self) -> bool:
        """Attempts to grab the I2C lock. Returns True on success."""
        ...
    def unlock(self) -> None:
        """Releases the I2C lock."""
        ...
    def readfrom_into(
        self,
        address: int,
        buffer: WriteableBuffer,
        *,
        start: int = 0,
        end: Optional[int] = None
    ) -> None:
        """Read into ``buffer`` from the device selected by ``address``.
        The number of bytes read will be the length of ``buffer``.
        At least one byte must be read.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]``. This will not cause an allocation like
        ``buf[start:end]`` will so it saves memory.

        :param int address: 7-bit device address
        :param ~_typing.WriteableBuffer buffer: buffer to write into
        :param int start: Index to start writing at
        :param int end: Index to write up to but not include"""
        ...
    def writeto(
        self,
        address: int,
        buffer: ReadableBuffer,
        *,
        start: int = 0,
        end: Optional[int] = None,
        stop: bool = True
    ) -> None:
        """Write the bytes from ``buffer`` to the device selected by ``address`` and then transmits a
        stop bit. Use `writeto_then_readfrom` when needing a write, no stop and repeated start
        before a read.

        If ``start`` or ``end`` is provided, then the buffer will be sliced
        as if ``buffer[start:end]``. This will not cause an allocation like
        ``buffer[start:end]`` will so it saves memory.

        Writing a buffer or slice of length zero is permitted, as it can be used
        to poll for the existence of a device.

        :param int address: 7-bit device address
        :param ~_typing.ReadableBuffer buffer: buffer containing the bytes to write
        :param int start: Index to start writing from
        :param int end: Index to read up to but not include"""
        ...
    def writeto_then_readfrom(
        self,
        address: int,
        out_buffer: ReadableBuffer,
        in_buffer: ReadableBuffer,
        *,
        out_start: int = 0,
        out_end: Optional[int] = None,
        in_start: int = 0,
        in_end: Optional[int] = None
    ) -> None:
        """Write the bytes from ``out_buffer`` to the device selected by ``address``, generate no stop
        bit, generate a repeated start and read into ``in_buffer``. ``out_buffer`` and
        ``in_buffer`` can be the same buffer because they are used sequentially.

        If ``start`` or ``end`` is provided, then the corresponding buffer will be sliced
        as if ``buffer[start:end]``. This will not cause an allocation like ``buf[start:end]``
        will so it saves memory.

        :param int address: 7-bit device address
        :param ~_typing.ReadableBuffer out_buffer: buffer containing the bytes to write
        :param ~_typing.WriteableBuffer in_buffer: buffer to write into
        :param int out_start: Index to start writing from
        :param int out_end: Index to read up to but not include. Defaults to ``len(buffer)``
        :param int in_start: Index to start writing at
        :param int in_end: Index to write up to but not include. Defaults to ``len(buffer)``"""

class OneWire:
    """Lowest-level of the Maxim OneWire protocol

    :class:`~bitbangio.OneWire` implements the timing-sensitive foundation of
    the Maxim (formerly Dallas Semi) OneWire protocol.

    Protocol definition is here: https://www.maximintegrated.com/en/app-notes/index.mvp/id/126"""

    def __init__(self, pin: microcontroller.Pin) -> None:

        """Create a OneWire object associated with the given pin. The object
        implements the lowest level timing-sensitive bits of the protocol.

        :param ~microcontroller.Pin pin: Pin to read pulses from.

        Read a short series of pulses::

          import bitbangio
          import board

          onewire = bitbangio.OneWire(board.D7)
          onewire.reset()
          onewire.write_bit(True)
          onewire.write_bit(False)
          print(onewire.read_bit())"""
        ...
    def deinit(self) -> None:
        """Deinitialize the OneWire bus and release any hardware resources for reuse."""
        ...
    def __enter__(self) -> OneWire:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def reset(self) -> bool:
        """Reset the OneWire bus"""
        ...
    def read_bit(self) -> bool:
        """Read in a bit

        :returns: bit state read
        :rtype: bool"""
        ...
    def write_bit(self, value: bool) -> None:
        """Write out a bit based on value."""
        ...

class SPI:
    """A 3-4 wire serial protocol

    SPI is a serial protocol that has exclusive pins for data in and out of the
    main device.  It is typically faster than :py:class:`~bitbangio.I2C` because a
    separate pin is used to select a device rather than a transmitted
    address. This class only manages three of the four SPI lines: `!clock`,
    `!MOSI`, `!MISO`. Its up to the client to manage the appropriate
    select line, often abbreviated `!CS` or `!SS`. (This is common because
    multiple secondaries can share the `!clock`, `!MOSI` and `!MISO` lines
    and therefore the hardware.)"""

    def __init__(
        self,
        clock: microcontroller.Pin,
        MOSI: Optional[microcontroller.Pin] = None,
        MISO: Optional[microcontroller.Pin] = None,
    ) -> None:
        """Construct an SPI object on the given pins.

        .. seealso:: Using this class directly requires careful lock management.
            Instead, use :class:`~adafruit_bus_device.spi_device.SPIDevice` to
            manage locks.

        .. seealso:: Using this class to directly read registers requires manual
            bit unpacking. Instead, use an existing driver or make one with
            :ref:`Register <register-module-reference>` data descriptors.


        :param ~microcontroller.Pin clock: the pin to use for the clock.
        :param ~microcontroller.Pin MOSI: the Main Out Selected In pin.
        :param ~microcontroller.Pin MISO: the Main In Selected Out pin."""
        ...
    def deinit(self) -> None:
        """Turn off the SPI bus."""
        ...
    def __enter__(self) -> SPI:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def configure(
        self,
        *,
        baudrate: int = 100000,
        polarity: int = 0,
        phase: int = 0,
        bits: int = 8
    ) -> None:
        """Configures the SPI bus. Only valid when locked.

        :param int baudrate: the clock rate in Hertz
        :param int polarity: the base state of the clock line (0 or 1)
        :param int phase: the edge of the clock that data is captured. First (0)
          or second (1). Rising or falling depends on clock polarity.
        :param int bits: the number of bits per word"""
        ...
    def try_lock(self) -> bool:
        """Attempts to grab the SPI lock. Returns True on success.

        :return: True when lock has been grabbed
        :rtype: bool"""
        ...
    def unlock(self) -> None:
        """Releases the SPI lock."""
        ...
    def write(self, buf: ReadableBuffer) -> None:
        """Write the data contained in ``buf``. Requires the SPI being locked.
        If the buffer is empty, nothing happens."""
        ...
    def readinto(
        self,
        buffer: WriteableBuffer,
        *,
        start: int = 0,
        end: Optional[int] = None,
        write_value: int = 0
    ) -> None:
        """Read into ``buffer`` while writing ``write_value`` for each byte read.
        The SPI object must be locked.
        If the number of bytes to read is 0, nothing happens.

        :param bytearray buffer: Read data into this buffer
        :param int start: Start of the slice of ``buffer`` to read into: ``buffer[start:end]``
        :param int end: End of the slice; this index is not included. Defaults to ``len(buffer)``
        :param int write_value: Value to write while reading."""
        ...
    def write_readinto(
        self,
        buffer_out: ReadableBuffer,
        buffer_in: ReadableBuffer,
        *,
        out_start: int = 0,
        out_end: Optional[int] = None,
        in_start: int = 0,
        in_end: Optional[int] = None
    ) -> None:
        """Write out the data in ``buffer_out`` while simultaneously reading data into ``buffer_in``.
        The SPI object must be locked.
        The lengths of the slices defined by ``buffer_out[out_start:out_end]`` and ``buffer_in[in_start:in_end]``
        must be equal.
        If buffer slice lengths are both 0, nothing happens.

        :param ~_typing.ReadableBuffer buffer_out: Write out the data in this buffer
        :param ~_typing.WriteableBuffer buffer_in: Read data into this buffer
        :param int out_start: Start of the slice of buffer_out to write out: ``buffer_out[out_start:out_end]``
        :param int out_end: End of the slice; this index is not included. Defaults to ``len(buffer_out)``
        :param int in_start: Start of the slice of ``buffer_in`` to read into: ``buffer_in[in_start:in_end]``
        :param int in_end: End of the slice; this index is not included. Defaults to ``len(buffer_in)``"""
        ...



from __future__ import annotations

from typing import Callable

def bisect(
    fun: Callable[[float], float],
    a: float,
    b: float,
    *,
    xtol: float = 2.4e-7,
    maxiter: int = 100
) -> float:
    """
    :param callable f: The function to bisect
    :param float a: The left side of the interval
    :param float b: The right side of the interval
    :param float xtol: The tolerance value
    :param float maxiter: The maximum number of iterations to perform

    Find a solution (zero) of the function ``f(x)`` on the interval
    (``a``..``b``) using the bisection method.  The result is accurate to within
    ``xtol`` unless more than ``maxiter`` steps are required."""
    ...

def fmin(
    fun: Callable[[float], float],
    x0: float,
    *,
    xatol: float = 2.4e-7,
    fatol: float = 2.4e-7,
    maxiter: int = 200
) -> float:
    """
    :param callable f: The function to bisect
    :param float x0: The initial x value
    :param float xatol: The absolute tolerance value
    :param float fatol: The relative tolerance value

    Find a minimum of the function ``f(x)`` using the downhill simplex method.
    The located ``x`` is within ``fxtol`` of the actual minimum, and ``f(x)``
    is within ``fatol`` of the actual minimum unless more than ``maxiter``
    steps are requried."""
    ...

def newton(
    fun: Callable[[float], float],
    x0: float,
    *,
    xtol: float = 2.4e-7,
    rtol: float = 0.0,
    maxiter: int = 50
) -> float:
    """
    :param callable f: The function to bisect
    :param float x0: The initial x value
    :param float xtol: The absolute tolerance value
    :param float rtol: The relative tolerance value
    :param float maxiter: The maximum number of iterations to perform

    Find a solution (zero) of the function ``f(x)`` using Newton's Method.
    The result is accurate to within ``xtol * rtol * |f(x)|`` unless more than
    ``maxiter`` steps are requried."""
    ...



"""Numerical and Statistical functions

Most of these functions take an "axis" argument, which indicates whether to
operate over the flattened array (None), or a particular axis (integer)."""

from __future__ import annotations

from typing import Optional, Union

import ulab
from ulab import _ArrayLike

def argmax(array: _ArrayLike, *, axis: Optional[int] = None) -> int:
    """Return the index of the maximum element of the 1D array"""
    ...

def argmin(array: _ArrayLike, *, axis: Optional[int] = None) -> int:
    """Return the index of the minimum element of the 1D array"""
    ...

def argsort(array: ulab.ndarray, *, axis: int = -1) -> ulab.ndarray:
    """Returns an array which gives indices into the input array from least to greatest."""
    ...

def cross(a: ulab.ndarray, b: ulab.ndarray) -> ulab.ndarray:
    """Return the cross product of two vectors of length 3"""
    ...

def diff(array: ulab.ndarray, *, n: int = 1, axis: int = -1) -> ulab.ndarray:
    """Return the numerical derivative of successive elements of the array, as
    an array.  axis=None is not supported."""
    ...

def flip(array: ulab.ndarray, *, axis: Optional[int] = None) -> ulab.ndarray:
    """Returns a new array that reverses the order of the elements along the
    given axis, or along all axes if axis is None."""
    ...

def max(array: _ArrayLike, *, axis: Optional[int] = None) -> float:
    """Return the maximum element of the 1D array"""
    ...

def mean(array: _ArrayLike, *, axis: Optional[int] = None) -> float:
    """Return the mean element of the 1D array, as a number if axis is None, otherwise as an array."""
    ...

def median(array: ulab.ndarray, *, axis: int = -1) -> ulab.ndarray:
    """Find the median value in an array along the given axis, or along all axes if axis is None."""
    ...

def min(array: _ArrayLike, *, axis: Optional[int] = None) -> float:
    """Return the minimum element of the 1D array"""
    ...

def roll(array: ulab.ndarray, distance: int, *, axis: Optional[int] = None) -> None:
    """Shift the content of a vector by the positions given as the second
    argument. If the ``axis`` keyword is supplied, the shift is applied to
    the given axis.  The array is modified in place."""
    ...

def sort(array: ulab.ndarray, *, axis: int = -1) -> ulab.ndarray:
    """Sort the array along the given axis, or along all axes if axis is None.
    The array is modified in place."""
    ...

def std(array: _ArrayLike, *, axis: Optional[int] = None, ddof: int = 0) -> float:
    """Return the standard deviation of the array, as a number if axis is None, otherwise as an array."""
    ...

def sum(
    array: _ArrayLike, *, axis: Optional[int] = None
) -> Union[float, int, ulab.ndarray]:
    """Return the sum of the array, as a number if axis is None, otherwise as an array."""
    ...



"""Support for "Parallel capture" interfaces"""

from __future__ import annotations

from typing import List, Optional

import microcontroller
from _typing import WriteableBuffer

class ParallelImageCapture:
    """Capture image frames from a camera with parallel data interface"""

    def __init__(
        self,
        *,
        data_pins: List[microcontroller.Pin],
        clock: microcontroller.Pin,
        vsync: Optional[microcontroller.Pin],
        href: Optional[microcontroller.Pin],
    ) -> None:
        """Create a parallel image capture object

        :param List[microcontroller.Pin] data_pins: The data pins.
        :param microcontroller.Pin clock: The pixel clock input.
        :param microcontroller.Pin vsync: The vertical sync input, which has a negative-going pulse at the beginning of each frame.
        :param microcontroller.Pin href: The horizontal reference input, which is high whenever the camera is transmitting valid pixel information.
        """
        ...
    def capture(
        self, buffer: WriteableBuffer, width: int, height: int, bpp: int = 16
    ) -> None:
        """Capture a single frame into the given buffer"""
        ...
    def deinit(self) -> None:
        """Deinitialize this instance"""
        ...
    def __enter__(self) -> ParallelImageCapture:
        """No-op used in Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware on context exit. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...



from __future__ import annotations

from typing import Union

import ulab

def dot(
    m1: ulab.numpy.ndarray, m2: ulab.numpy.ndarray
) -> Union[ulab.numpy.ndarray, float]:
    """
    :param ~ulab.numpy.ndarray m1: a matrix, or a vector
    :param ~ulab.numpy.ndarray m2: a matrix, or a vector

    Computes the product of two matrices, or two vectors. In the letter case, the inner product is returned."""
    ...



"""Support for MP3-compressed audio files"""

from __future__ import annotations

import typing

from _typing import WriteableBuffer

class MP3Decoder:
    """Load a mp3 file for audio playback"""

    def __init__(self, file: typing.BinaryIO, buffer: WriteableBuffer) -> None:

        """Load a .mp3 file for playback with `audioio.AudioOut` or `audiobusio.I2SOut`.

        :param typing.BinaryIO file: Already opened mp3 file
        :param ~_typing.WriteableBuffer buffer: Optional pre-allocated buffer, that will be split in half and used for double-buffering of the data. If not provided, two buffers are allocated internally.  The specific buffer size required depends on the mp3 file.


        Playing a mp3 file from flash::

          import board
          import audiomp3
          import audioio
          import digitalio

          # Required for CircuitPlayground Express
          speaker_enable = digitalio.DigitalInOut(board.SPEAKER_ENABLE)
          speaker_enable.switch_to_output(value=True)

          data = open("cplay-16bit-16khz-64kbps.mp3", "rb")
          mp3 = audiomp3.MP3Decoder(data)
          a = audioio.AudioOut(board.A0)

          print("playing")
          a.play(mp3)
          while a.playing:
            pass
          print("stopped")"""
        ...
    def deinit(self) -> None:
        """Deinitialises the MP3 and releases all memory resources for reuse."""
        ...
    def __enter__(self) -> MP3Decoder:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    file: typing.BinaryIO
    """File to play back."""

    sample_rate: int
    """32 bit value that dictates how quickly samples are loaded into the DAC
    in Hertz (cycles per second). When the sample is looped, this can change
    the pitch output without changing the underlying sample."""

    bits_per_sample: int
    """Bits per sample. (read only)"""

    channel_count: int
    """Number of audio channels. (read only)"""

    rms_level: float
    """The RMS audio level of a recently played moment of audio. (read only)"""



"""TCP, UDP and RAW socket support

.. warning:: This module is disabled in 6.x and will removed in 7.x. Please use networking
             libraries instead. (Native networking will provide a socket compatible class.)

Create TCP, UDP and RAW sockets for communicating over the Internet."""

from __future__ import annotations

from typing import Optional, Tuple

from _typing import ReadableBuffer, WriteableBuffer

class socket:

    AF_INET: int
    AF_INET6: int
    SOCK_STREAM: int
    SOCK_DGRAM: int
    SOCK_RAW: int
    IPPROTO_TCP: int
    def __init__(
        self, family: int = AF_INET, type: int = SOCK_STREAM, proto: int = IPPROTO_TCP
    ) -> None:
        """Create a new socket

        :param int family: AF_INET or AF_INET6
        :param int type: SOCK_STREAM, SOCK_DGRAM or SOCK_RAW
        :param int proto: IPPROTO_TCP, IPPROTO_UDP or IPPROTO_RAW (ignored)"""
        ...
    def bind(self, address: Tuple[str, int]) -> None:
        """Bind a socket to an address

        :param address: tuple of (remote_address, remote_port)
        :type address: tuple(str, int)"""
        ...
    def listen(self, backlog: int) -> None:
        """Set socket to listen for incoming connections

        :param int backlog: length of backlog queue for waiting connetions"""
        ...
    def accept(self) -> Tuple[socket, str]:
        """Accept a connection on a listening socket of type SOCK_STREAM,
        creating a new socket of type SOCK_STREAM.
        Returns a tuple of (new_socket, remote_address)"""
    def connect(self, address: Tuple[str, int]) -> None:
        """Connect a socket to a remote address

        :param address: tuple of (remote_address, remote_port)
        :type address: tuple(str, int)"""
        ...
    def send(self, bytes: ReadableBuffer) -> int:
        """Send some bytes to the connected remote address.
        Suits sockets of type SOCK_STREAM

        :param ~_typing.ReadableBuffer bytes: some bytes to send"""
        ...
    def recv_into(self, buffer: WriteableBuffer, bufsize: int) -> int:
        """Reads some bytes from the connected remote address, writing
        into the provided buffer. If bufsize <= len(buffer) is given,
        a maximum of bufsize bytes will be read into the buffer. If no
        valid value is given for bufsize, the default is the length of
        the given buffer.

        Suits sockets of type SOCK_STREAM
        Returns an int of number of bytes read.

        :param ~_typing.WriteableBuffer buffer: buffer to receive into
        :param int bufsize: optionally, a maximum number of bytes to read."""
        ...
    def recv(self, bufsize: int) -> bytes:
        """Reads some bytes from the connected remote address.
        Suits sockets of type SOCK_STREAM
        Returns a bytes() of length <= bufsize

        :param int bufsize: maximum number of bytes to receive"""
        ...
    def sendto(self, bytes: ReadableBuffer, address: Tuple[str, int]) -> int:
        """Send some bytes to a specific address.
        Suits sockets of type SOCK_DGRAM

        :param ~_typing.ReadableBuffer bytes: some bytes to send
        :param address: tuple of (remote_address, remote_port)
        :type address: tuple(str, int)"""
        ...
    def recvfrom(self, bufsize: int) -> Tuple[bytes, Tuple[str, int]]:
        """Reads some bytes from the connected remote address.
        Suits sockets of type SOCK_STREAM

        Returns a tuple containing
        * a bytes() of length <= bufsize
        * a remote_address, which is a tuple of ip address and port number

        :param int bufsize: maximum number of bytes to receive"""
        ...
    def setsockopt(self, level: int, optname: int, value: int) -> None:
        """Sets socket options"""
        ...
    def settimeout(self, value: int) -> None:
        """Set the timeout value for this socket.

        :param int value: timeout in seconds.  0 means non-blocking.  None means block indefinitely."""
        ...
    def setblocking(self, flag: bool) -> Optional[int]:
        """Set the blocking behaviour of this socket.

        :param bool flag: False means non-blocking, True means block indefinitely."""
        ...

def getaddrinfo(host: str, port: int) -> Tuple[int, int, int, str, str]:
    """Gets the address information for a hostname and port

    Returns the appropriate family, socket type, socket protocol and
    address information to call socket.socket() and socket.connect() with,
    as a tuple."""
    ...
